'''
See COPYRIGHT.md for copyright information.
'''
from arelle import XbrlConst, XmlUtil
from arelle.formula import XPathContext
from arelle.ModelFormulaObject import (ModelTypedDimension, ModelParameter)
from arelle.Aspect import (aspectModels, aspectStr, Aspect)
from arelle.ModelRenderingObject import (DefnMdlDefinitionNode,
                                         DefnMdlBreakdown,
                                         DefnMdlClosedDefinitionNode,
                                         DefnMdlRuleDefinitionNode,
                                         DefnMdlAspectNode,
                                         DefnMdlRelationshipNode,
                                         DefnMdlDimensionRelationshipNode)
from arelle.ModelValue import (QName)

def init(modelXbrl):
    # setup modelXbrl for rendering evaluation

    # dimension defaults required in advance of validation
    from arelle import ValidateXbrlDimensions, ModelDocument
    from arelle.formula import ValidateFormula, FormulaEvaluator
    ValidateXbrlDimensions.loadDimensionDefaults(modelXbrl)

    hasXbrlTables = False

    # validate table linkbase dimensions
    for baseSetKey in modelXbrl.baseSets.keys():
        arcrole, ELR, linkqname, arcqname = baseSetKey
        if ELR and linkqname and arcqname and XbrlConst.isTableRenderingArcrole(arcrole):
            ValidateFormula.checkBaseSet(modelXbrl, arcrole, ELR, modelXbrl.relationshipSet(arcrole,ELR,linkqname,arcqname))
            if arcrole in (XbrlConst.tableBreakdown, XbrlConst.tableBreakdownMMDD):
                hasXbrlTables = True

    # provide context for view
    if modelXbrl.modelDocument.type in (ModelDocument.Type.INSTANCE, ModelDocument.Type.INLINEXBRL, ModelDocument.Type.INLINEXBRLDOCUMENTSET):
        instance = None # use instance of the entry point
    else: # need dummy instance
        instance = ModelDocument.create(modelXbrl, ModelDocument.Type.INSTANCE,
                                        "dummy.xml",  # fake URI and fake schemaRef
                                        ("http://www.xbrl.org/2003/xbrl-instance-2003-12-31.xsd",))

    if hasXbrlTables:
        FormulaEvaluator.init()
        modelXbrl.rendrCntx = XPathContext.create(modelXbrl, instance)

        modelXbrl.profileStat(None)

        # setup fresh parameters from formula options
        modelXbrl.parameters = modelXbrl.modelManager.formulaOptions.typedParameters(modelXbrl.prefixedNamespaces)

        # validate parameters and custom function signatures
        ValidateFormula.validate(modelXbrl, xpathContext=modelXbrl.rendrCntx, parametersOnly=True, statusMsg=_("compiling rendering tables"))

        # compile and validate tables
        for modelTable in modelXbrl.modelRenderingTables:
            modelTable.fromInstanceQnames = None # required if referred to by variables scope chaining
            modelTable.compile()
            # remove unwanted messages when running conformance suite
            if (len(modelXbrl.factsInInstance) == 0 and # these errors not expected when there are no instance facts
                modelXbrl.modelManager.loadedModelXbrls[0].modelDocument.type in ModelDocument.Type.TESTCASETYPES):
                for i in range(len(modelXbrl.errors)):
                    if modelXbrl.errors[i] in ("xfie:invalidExplicitDimensionQName",):
                        del modelXbrl.errors[i]

            modelTable.priorAspectAxisDisposition = {}
            # check ordinate aspects against aspectModel
            oppositeAspectModel = ({'dimensional','non-dimensional'} - {modelTable.aspectModel}).pop()
            uncoverableAspects = ()
            aspectsCovered = set()
            for tblBrkdnRel in modelXbrl.relationshipSet((XbrlConst.tableBreakdown, XbrlConst.tableBreakdownMMDD)).fromModelObject(modelTable):
                breakdownAspectsCovered = set()
                hasCoveredAspect = checkBreakdownDefinitionNode(modelXbrl, modelTable, tblBrkdnRel, tblBrkdnRel.axis, uncoverableAspects, breakdownAspectsCovered)
                aspectsCovered |= breakdownAspectsCovered
                checkBreakdownLeafNodeAspects(modelXbrl, modelTable, tblBrkdnRel, set(), breakdownAspectsCovered)
            if Aspect.CONCEPT not in aspectsCovered:
                modelXbrl.error("xbrlte:tableMissingConceptAspect",
                    _("Table %(xlinkLabel)s does not include the concept aspect as one of its participating aspects"),
                    modelObject=modelTable, xlinkLabel=modelTable.xlinkLabel)
            del modelTable.priorAspectAxisDisposition
            # check for table-parameter name clash
            parameterNames = {}
            for tblParamRel in modelXbrl.relationshipSet((XbrlConst.tableParameter, XbrlConst.tableParameterMMDD)).fromModelObject(modelTable):
                parameterName = tblParamRel.variableQname
                if parameterName in parameterNames:
                    modelXbrl.error("xbrlte:tableParameterNameClash",
                        _("Table %(xlinkLabel)s has parameter name clash for variable %(name)s"),
                        modelObject=(modelTable,tblParamRel,parameterNames[parameterName]), xlinkLabel=modelTable.xlinkLabel, name=parameterName)
                else:
                    parameterNames[parameterName] = tblParamRel

        if instance is not None: # no instance was provided, check for context-dependent XPath expressions
            for paramQname, modelParameter in modelXbrl.qnameParameters.items():
                if isinstance(modelParameter, ModelParameter):
                    if any(hasattr(p, "name") and p.name in XPathContext.PATH_OPS for p in modelParameter.selectProg):
                        modelXbrl.error("arelle:tableParameterRequiresInstance",
                            _("Parameter %(qname)s requires an instance as a context item for the XPath expression but no instance was provided."),
                            modelObject=modelParameter, qname=paramQname)

        modelXbrl.profileStat(_("compileTables"))

def checkBreakdownDefinitionNode(modelXbrl, modelTable, tblBrkdnRel, tblAxisDisposition, uncoverableAspects, aspectsCovered):
    definitionNode = tblBrkdnRel.toModelObject
    hasCoveredAspect = False
    if isinstance(definitionNode, DefnMdlDefinitionNode):
        for aspect in definitionNode.aspectsCovered():
            aspectsCovered.add(aspect)
            if (aspect in uncoverableAspects or
                (isinstance(aspect, QName) and modelTable.aspectModel == 'non-dimensional')):
                modelXbrl.error("xbrlte:axisAspectModelMismatch",
                    _("%(definitionNode)s %(xlinkLabel)s, aspect model %(aspectModel)s, aspect %(aspect)s not allowed"),
                    modelObject=modelTable, definitionNode=definitionNode.localName, xlinkLabel=definitionNode.xlinkLabel, aspectModel=modelTable.aspectModel,
                    aspect=str(aspect) if isinstance(aspect,QName) else Aspect.label[aspect])
            hasCoveredAspect = True
            if aspect in modelTable.priorAspectAxisDisposition:
                otherAxisDisposition, otherDefinitionNode = modelTable.priorAspectAxisDisposition[aspect]
                if tblAxisDisposition != otherAxisDisposition and aspect != Aspect.DIMENSIONS:
                    modelXbrl.error("xbrlte:aspectClashBetweenBreakdowns",
                        _("%(definitionNode)s %(xlinkLabel)s, aspect %(aspect)s defined on axes of disposition %(axisDisposition)s and %(axisDisposition2)s"),
                        modelObject=(modelTable, definitionNode, otherDefinitionNode), definitionNode=definitionNode.localName, xlinkLabel=definitionNode.xlinkLabel,
                        axisDisposition=tblAxisDisposition, axisDisposition2=otherAxisDisposition,
                        aspect=str(aspect) if isinstance(aspect,QName) else Aspect.label[aspect])
            else:
                modelTable.priorAspectAxisDisposition[aspect] = (tblAxisDisposition, definitionNode)
        ruleSetChildren = XmlUtil.children(definitionNode, definitionNode.namespaceURI, "ruleSet")
        if definitionNode.isMerged or isinstance(definitionNode, (DefnMdlRelationshipNode, DefnMdlAspectNode)):
            labelRels = modelXbrl.relationshipSet(XbrlConst.elementLabel).fromModelObject(definitionNode)
            if labelRels:
                modelXbrl.error("xbrlte:invalidUseOfLabel",
                    _("Merged %(definitionNode)s %(xlinkLabel)s has label(s)"),
                    modelObject=[modelTable, definitionNode] + [r.toModelObject for r in labelRels],
                    definitionNode=definitionNode.localName, xlinkLabel=definitionNode.xlinkLabel)
        if definitionNode.isMerged:
            if ruleSetChildren:
                modelXbrl.error("xbrlte:mergedRuleNodeWithTaggedRuleSet",
                    _("Merged %(definitionNode)s %(xlinkLabel)s has tagged rule set(s)"),
                    modelObject=[modelTable, definitionNode] + ruleSetChildren,
                    definitionNode=definitionNode.localName, xlinkLabel=definitionNode.xlinkLabel)
            if not definitionNode.isAbstract:
                modelXbrl.error("xbrlte:nonAbstractMergedRuleNode",
                    _("Merged %(definitionNode)s %(xlinkLabel)s is not abstract"),
                    modelObject=(modelTable, definitionNode), definitionNode=definitionNode.localName, xlinkLabel=definitionNode.xlinkLabel)
    if isinstance(definitionNode, DefnMdlRuleDefinitionNode):
        tagConstraintSets = {}
        otherConstraintSet = None
        # must look at xml constructs for duplicates
        for ruleSet in XmlUtil.children(definitionNode, definitionNode.namespaceURI, "ruleSet"):
            tag = ruleSet.tagName
            if tag is not None: # named constraint sets only
                for aspect in ruleSet.aspectsCovered():
                    if aspect != Aspect.DIMENSIONS:
                        modelTable.aspectsInTaggedConstraintSets.add(aspect)
            if tag in tagConstraintSets:
                modelXbrl.error("xbrlte:duplicateTag",
                    _("%(definitionNode)s %(xlinkLabel)s duplicate rule set tags %(tag)s"),
                    modelObject=(modelTable, definitionNode, tagConstraintSets[tag], ruleSet),
                    definitionNode=definitionNode.localName, xlinkLabel=definitionNode.xlinkLabel, tag=tag)
            else:
                tagConstraintSets[tag] = ruleSet
            if aspect in ruleSet.aspectsCovered():
                hasCoveredAspect = True
                aspectsCovered.add(aspect)
        for tag, constraintSet in definitionNode.constraintSets.items():
            if otherConstraintSet is None:
                otherConstraintSet = constraintSet
            elif otherConstraintSet.aspectsModelCovered() != constraintSet.aspectsModelCovered():
                modelXbrl.error("xbrlte:constraintSetAspectMismatch",
                    _("%(definitionNode)s %(xlinkLabel)s constraint set mismatches between %(tag1)s and %(tag2)s in constraints %(aspects)s"),
                    modelObject=(modelTable, definitionNode, otherConstraintSet, constraintSet),
                    definitionNode=definitionNode.localName, xlinkLabel=definitionNode.xlinkLabel,
                    tag1=getattr(otherConstraintSet,"tagName","(no tag)"), tag2=getattr(constraintSet, "tagName", "(no tag)"),
                    aspects=", ".join(aspectStr(aspect)
                                      for aspect in otherConstraintSet.aspectsCovered() ^ constraintSet.aspectsCovered()
                                      if aspect != Aspect.DIMENSIONS))
    if isinstance(definitionNode, DefnMdlDimensionRelationshipNode):
        hasCoveredAspect = True
        if modelTable.aspectModel == 'non-dimensional':
            modelXbrl.error("xbrlte:axisAspectModelMismatch",
                _("DimensionRelationship axis %(xlinkLabel)s can't be used in non-dimensional aspect model"),
                modelObject=(modelTable,definitionNode), xlinkLabel=definitionNode.xlinkLabel)
        # TLB 1.0 legacy use of Hypercube Linkroles to find domain roots
        if modelXbrl.parameters.get(QName(None,None,"tlbDimRelsUseHcRoleForDomainRoots"),("",""))[1] in ("true", True):
            definitionNode.tlbDimRelsUseHcRoleForDomainRoots = True

    definitionNodeHasChild = False
    for axisSubtreeRel in modelXbrl.relationshipSet((XbrlConst.tableBreakdownTree, XbrlConst.tableBreakdownTreeMMDD, XbrlConst.tableDefinitionNodeSubtree, XbrlConst.tableDefinitionNodeSubtreeMMDD)).fromModelObject(definitionNode):
        if checkBreakdownDefinitionNode(modelXbrl, modelTable, axisSubtreeRel, tblAxisDisposition, uncoverableAspects, aspectsCovered):
            hasCoveredAspect = True # something below was covering
        definitionNodeHasChild = True
    if isinstance(definitionNode, DefnMdlAspectNode):
        for aspect in definitionNode.aspectsCovered():
            if isinstance(aspect, QName): # dimension aspect
                concept = modelXbrl.qnameConcepts.get(aspect)
                if concept is None or not concept.isDimensionItem:
                    modelXbrl.error("xbrlte:invalidDimensionQNameOnAspectNode",
                        _("Aspect node %(xlinkLabel)s dimensional aspect %(dimension)s is not a dimension"),
                        modelObject=(modelTable,definitionNode), xlinkLabel=definitionNode.xlinkLabel, dimension=aspect)
            for rel in definitionNode.filterRelationships:
                filter = rel.toModelObject
                if filter is not None:
                    if isinstance(filter, ModelTypedDimension) and filter.dimQname:
                        concept = modelXbrl.qnameConcepts.get(filter.dimQname)
                        if (concept is None or not concept.isDimensionItem) and len(modelXbrl.factsInInstance) > 0:
                            modelXbrl.error("xfie:invalidTypedDimensionQName",
                                _("Aspect node %(xlinkLabel)s dimensional aspect %(dimension)s is not a dimension"),
                                modelObject=(modelTable,definitionNode), xlinkLabel=definitionNode.xlinkLabel, dimension=aspect)

    if not definitionNodeHasChild:
        if (definitionNode.namespaceURI in ("http://www.eurofiling.info/2010/rendering", "http://xbrl.org/2011/table")
            and not hasCoveredAspect):
            modelXbrl.error("xbrlte:aspectValueNotDefinedByOrdinate",
                _("%(definitionNode)s %(xlinkLabel)s does not define an aspect"),
                modelObject=(modelTable,definitionNode), xlinkLabel=definitionNode.xlinkLabel, definitionNode=definitionNode.localName)
        if (isinstance(definitionNode, DefnMdlClosedDefinitionNode) and
            definitionNode.isAbstract):
            modelXbrl.error("xbrlte:abstractRuleNodeNoChildren",
                _("Abstract %(definitionNode)s %(xlinkLabel)s has no children"),
                modelObject=(modelTable,definitionNode), xlinkLabel=definitionNode.xlinkLabel, definitionNode=definitionNode.localName)
    return hasCoveredAspect

def checkBreakdownLeafNodeAspects(modelXbrl, modelTable, tblBrkdnRel, parentAspectsCovered, breakdownAspects):
    breakdown = tblBrkdnRel.toModelObject

    for dfnRel in modelXbrl.relationshipSet((XbrlConst.tableBreakdownTree,XbrlConst.tableBreakdownTreeMMDD)).fromModelObject(breakdown):
        definitionNode = dfnRel.toModelObject
        aspectsCovered = parentAspectsCovered.copy()
        if isinstance(definitionNode, DefnMdlDefinitionNode):
            for aspect in definitionNode.aspectsCovered():
                aspectsCovered.add(aspect)
            if isinstance(definitionNode, DefnMdlRuleDefinitionNode):
                for ruleSet in XmlUtil.children(definitionNode, definitionNode.namespaceURI, "ruleSet"):
                    for aspect in ruleSet.aspectsCovered():
                        aspectsCovered.add(aspect)
            definitionNodeHasChild = False
            for axisSubtreeRel in modelXbrl.relationshipSet((XbrlConst.tableBreakdownTree, XbrlConst.tableBreakdownTreeMMDD, XbrlConst.tableDefinitionNodeSubtree, XbrlConst.tableDefinitionNodeSubtreeMMDD)).fromModelObject(definitionNode):
                checkBreakdownLeafNodeAspects(modelXbrl, modelTable, axisSubtreeRel, aspectsCovered, breakdownAspects)
                definitionNodeHasChild = True

            if not definitionNode.isAbstract and not isinstance(definitionNode, DefnMdlBreakdown): # this is a leaf node
                missingAspects = set(aspect
                                     for aspect in breakdownAspects
                                     if aspect not in aspectsCovered and
                                        aspect not in (Aspect.DIMENSIONS, Aspect.OMIT_DIMENSIONS) and not isinstance(aspect,QName))

                isForever = definitionNode.aspectValue(modelXbrl.rendrCntx, Aspect.PERIOD_TYPE) == "forever"
                # a definition node cannot define a period and a instant, but it's not the case for the breakdown Aspects
                if (Aspect.START in aspectsCovered and Aspect.END in aspectsCovered) or isForever:
                    missingAspects.discard(Aspect.INSTANT)
                elif Aspect.INSTANT in aspectsCovered or isForever:
                    missingAspects.discard(Aspect.START)
                    missingAspects.discard(Aspect.END)

                if (missingAspects):
                    modelXbrl.error("xbrlte:missingAspectValue",
                        _("%(definitionNode)s %(xlinkLabel)s does not define an aspect for %(aspect)s"),
                        modelObject=(modelTable,definitionNode), xlinkLabel=definitionNode.xlinkLabel, definitionNode=definitionNode.localName,
                        aspect=', '.join(aspectStr(aspect) for aspect in missingAspects))



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

import regex


def findProhibitedCharacters(
    text: str | None,
    pattern: regex.Pattern[str],
) -> set[str]:
    """
    Find all characters in text that match the prohibited-character pattern.

    A prohibited-character pattern typically matches characters NOT in an allowed set.

    :param text: Text to scan (None or empty returns empty set)
    :param pattern: Compiled regex that matches prohibited characters
    :return: Set of matched prohibited characters
    """
    if not text:
        return set()
    return set(pattern.findall(text))



ARL_2022_PREVIEW = "arl-2022-preview"
ARL_2024_PREVIEW = "arl-2024-preview"
ARL_2025_PREVIEW = "arl-2025-preview"
ARL_2024_MULTI_TARGET_PREVIEW = "arl-2024-multi-target-preview"
ARL_2025_MULTI_TARGET_PREVIEW = "arl-2025-multi-target-preview"
DKFIN_2024_MULTI_TARGET_PREVIEW = "dkfin-2024-multi-target-preview"

ALL_DISCLOSURE_SYSTEMS = [
    ARL_2022_PREVIEW,
    ARL_2024_PREVIEW,
    ARL_2025_PREVIEW,
    ARL_2024_MULTI_TARGET_PREVIEW,
    ARL_2025_MULTI_TARGET_PREVIEW,
    DKFIN_2024_MULTI_TARGET_PREVIEW
]

ARL_DISCLOSURE_SYSTEMS = [
    ARL_2022_PREVIEW,
    ARL_2024_PREVIEW,
    ARL_2025_PREVIEW,
    ARL_2024_MULTI_TARGET_PREVIEW,
    ARL_2025_MULTI_TARGET_PREVIEW
]

STAND_ALONE_DISCLOSURE_SYSTEMS = [
    ARL_2022_PREVIEW,
    ARL_2024_PREVIEW,
    ARL_2025_PREVIEW
]

ARL_MULTI_TARGET_DISCLOSURE_SYSTEMS = [
    ARL_2024_MULTI_TARGET_PREVIEW,
    ARL_2025_MULTI_TARGET_PREVIEW
]

DKFIN_MULTI_TARGET_DISCLOSURE_SYSTEMS = [
    DKFIN_2024_MULTI_TARGET_PREVIEW
]



'''
See COPYRIGHT.md for copyright information.
'''
import os, threading, time, logging, sys, traceback
from collections import OrderedDict, defaultdict
from tkinter import Menu, BooleanVar, font as tkFont
from arelle.ModelFormulaObject import Aspect, aspectModels, aspectModelAspect
from arelle import (ViewWinTkTable, ModelDtsObject, ModelInstanceObject, XbrlConst,
                    ModelXbrl, Locale, FunctionXfi,
                    ValidateXbrlDimensions, ViewFileRenderedGrid, ViewFileRenderedLayout, ViewFileRenderedStructure)
from arelle.ModelDocumentType import ModelDocumentType
from arelle.ModelValue import qname, QName
from arelle.rendering.RenderingResolution import RENDER_UNITS_PER_CHAR
from arelle.rendering.RenderingLayout import layoutTable
from arelle.ModelInstanceObject import ModelDimensionValue
from arelle.ModelRenderingObject import (StrctMdlBreakdown, DefnMdlDefinitionNode,
                                         DefnMdlClosedDefinitionNode, DefnMdlAspectNode,
                                         OPEN_ASPECT_ENTRY_SURROGATE)
from arelle.formula.FormulaEvaluator import init as formulaEvaluatorInit, aspectMatches

from arelle.PrototypeInstanceObject import FactPrototype
from arelle.UITkTable import XbrlTable
from arelle.DialogNewFactItem import getNewFactItemOptions
from collections import defaultdict
from arelle.ValidateXbrl import ValidateXbrl
from arelle.XbrlConst import eurofilingModelNamespace, eurofilingModelPrefix
from arelle.ValidateXbrlDimensions import isFactDimensionallyValid
from arelle.XmlValidate import UNVALIDATED, validate as xmlValidate

TRACE_TK = False # print trace messages of tk table interface

try:
    from tkinter import ttk
    _Combobox = ttk.Combobox
except ImportError:
    from ttk import Combobox
    _Combobox = Combobox

emptyList = []

ENTRY_WIDTH_IN_CHARS = 12 # width of a data column entry cell in characters (nominal)
ENTRY_WIDTH_SCREEN_UNITS = 100
PADDING = 20 # screen units of padding between entry cells

qnPercentItemType = qname("{http://www.xbrl.org/dtr/type/numeric}num:percentItemType")
qnPureItemType = qname("{http://www.xbrl.org/2003/instance}xbrli:pureItemType")
integerItemTypes = {"integerItemType", "nonPositiveIntegerItemType", "negativeIntegerItemType",
                    "longItemType", "intItemType", "shortItemType", "byteItemType",
                    "nonNegativeIntegerItemType", "unsignedLongItemType", "unsignedIntItemType",
                    "unsignedShortItemType", "unsignedByteItemType", "positiveIntegerItemType"}
TABLE_AXIS_ROLES = (XbrlConst.euTableAxis, XbrlConst.tableBreakdown, XbrlConst.tableBreakdownMMDD)

'''
Returns a tuple with all known table axis roles
'''
def getTableAxisArcroles():
    return TABLE_AXIS_ROLES

def viewRenderedGrid(modelXbrl, tabWin, lang=None):
    modelXbrl.modelManager.showStatus(_("viewing rendering"))
    view = ViewRenderedGrid(modelXbrl, tabWin, lang)
    if not view.table.isInitialized: # unable to load or initialize tktable
        return None

    modelXbrl.modelManager.addToLog(_("Loading experimental table linkbase viewer"))

    view.blockMenuEvents = 1

    menu = view.contextMenu()
    optionsMenu = Menu(view.viewFrame, tearoff=0)
    optionsMenu.add_command(label=_("New fact item options"), underline=0, command=lambda: getNewFactItemOptions(modelXbrl.modelManager.cntlr, view.newFactItemOptions))
    optionsMenu.add_command(label=_("Open breakdown entry rows"), underline=0, command=view.setOpenBreakdownEntryRows)
    view.ignoreDimValidity.trace("w", view.viewReloadDueToMenuAction)
    optionsMenu.add_checkbutton(label=_("Ignore Dimensional Validity"), underline=0, variable=view.ignoreDimValidity, onvalue=True, offvalue=False)
    menu.add_cascade(label=_("Options"), menu=optionsMenu, underline=0)
    view.tablesMenu = Menu(view.viewFrame, tearoff=0)
    menu.add_cascade(label=_("Tables"), menu=view.tablesMenu, underline=0)
    view.tablesMenuLength = 0
    view.menuAddLangs()
    saveMenu = Menu(view.viewFrame, tearoff=0)
    saveMenu.add_command(label=_("HTML table"), underline=0, command=lambda: view.modelXbrl.modelManager.cntlr.fileSave(
        view=view, fileType="html", method=ViewFileRenderedGrid.viewRenderedGrid, caption=_("arelle - Save HTML-rendered Table")))
    saveMenu.add_command(label=_("Excel table"), underline=0, command=lambda: view.modelXbrl.modelManager.cntlr.fileSave(
        view=view, fileType="xlsx", method=ViewFileRenderedGrid.viewRenderedGrid, caption=_("arelle - Save Excel-rendered Table")))
    saveMenu.add_command(label=_("Layout model"), underline=0, command=lambda: view.modelXbrl.modelManager.cntlr.fileSave(
        view=view, fileType="xml", method=ViewFileRenderedLayout.viewRenderedLayout, caption=_("arelle - Save Table Layout Model")))
    saveMenu.add_command(label=_("Structural model"), underline=0, command=lambda: view.modelXbrl.modelManager.cntlr.fileSave(
        view=view, fileType="json", method=ViewFileRenderedStructure.viewRenderedStructuralModel, caption=_("arelle - Save Table Structural Model")))
    saveMenu.add_command(label=_("XBRL instance"), underline=0, command=view.saveInstance)
    menu.add_cascade(label=_("Save"), menu=saveMenu, underline=0)
    view.view()
    view.blockSelectEvent = 1
    view.blockViewModelObject = 0
    view.viewFrame.bind("<Enter>", view.cellEnter, '+')
    view.viewFrame.bind("<Leave>", view.cellLeave, '+')
    view.viewFrame.bind("<FocusOut>", view.onQuitView, '+')
    view.viewFrame.bind("<1>", view.onClick, '+') # does not currently work (since tktable changes)
    view.viewFrame.bind("<Configure>", view.onConfigure, '+') # frame resized, redo column header wrap length ratios
    view.blockMenuEvents = 0
    if "saveTableStructuralModel" in modelXbrl.modelManager.formulaOptions.parameterValues:
        ViewFileRenderedStructure.viewRenderedStructuralModel(modelXbrl,
              modelXbrl.modelManager.formulaOptions.parameterValues["saveTableStructuralModel"][1],
              lang=lang, sourceView=view)
    if "saveTableLayoutModel" in modelXbrl.modelManager.formulaOptions.parameterValues:
        ViewFileRenderedLayout.viewRenderedLayout(modelXbrl,
              modelXbrl.modelManager.formulaOptions.parameterValues["saveTableLayoutModel"][1],
              lang=lang, sourceView=view)
    if "saveHtmlTable" in modelXbrl.modelManager.formulaOptions.parameterValues:
        ViewFileRenderedGrid.viewRenderedGrid(modelXbrl,
              modelXbrl.modelManager.formulaOptions.parameterValues["saveHtmlTable"][1],
              lang=lang, sourceView=view)
    if "saveTable" in modelXbrl.modelManager.formulaOptions.parameterValues:
        ViewFileRenderedGrid.viewRenderedGrid(modelXbrl,
              modelXbrl.modelManager.formulaOptions.parameterValues["saveTable"][1],
              lang=lang, sourceView=view)
    return view

class ViewRenderedGrid(ViewWinTkTable.ViewTkTable):
    def __init__(self, modelXbrl, tabWin, lang):
        super(ViewRenderedGrid, self).__init__(modelXbrl, tabWin, _("Table"),
                                               False, lang, self.onQuitView)
        self.newFactItemOptions = ModelInstanceObject.NewFactItemOptions(xbrlInstance=modelXbrl)
        self.factPrototypes = []
        self.aspectEntryObjectIdsNode = {}
        self.aspectEntryObjectIdsCell = {}
        self.factPrototypeAspectEntryObjectIds = defaultdict(set)
        self.zHdrElts = None
        # context menu Boolean vars
        self.options = self.modelXbrl.modelManager.cntlr.config.setdefault("viewRenderedGridOptions", {})
        self.openBreakdownLines = self.options.setdefault("openBreakdownLines", 5) # ensure there is a default entry
        self.ignoreDimValidity = BooleanVar(value=self.options.setdefault("ignoreDimValidity",True))
        formulaEvaluatorInit() # one-time module initialization
        self.conceptMessageIssued = False
        self.tblMenuEntries = {}

    def close(self):
        super(ViewRenderedGrid, self).close()
        if self.modelXbrl:
            for fp in self.factPrototypes:
                fp.clear()
            self.factPrototypes = None
            self.aspectEntryObjectIdsNode.clear()
            self.aspectEntryObjectIdsCell.clear()
            self.rendrCntx = None # remove the reference but do not manipulate since it may still be in use and shared

    def loadTablesMenu(self):
        if not self.tblMenuEntries:
            self.tablesToELR = {}
            for lytMdlTableSet in self.lytMdlTblMdl.lytMdlTableSets:
                # table name
                modelRoleTypes = self.modelXbrl.roleTypes.get(lytMdlTableSet.srcLinkrole)
                if modelRoleTypes is not None and len(modelRoleTypes) > 0:
                    # roledefinition = modelRoleTypes[0].definition
                    roledefinition = self.modelXbrl.roleTypeDefinition(lytMdlTableSet.srcLinkrole, self.lang) # Definition in selected language
                    if roledefinition is None or roledefinition == "":
                        roledefinition = os.path.basename(lytMdlTableSet.srcLinkrole)
                    # add table to menu if there's any entry
                    self.tblMenuEntries[roledefinition] = lytMdlTableSet.srcLinkrole
                    # find table group object corresponding to this table for EBA filings from the root table object
                    for tableAxisArcrole in getTableAxisArcroles():
                        tblAxisRelSet = self.modelXbrl.relationshipSet(tableAxisArcrole, lytMdlTableSet.srcLinkrole)
                        if tblAxisRelSet and len(tblAxisRelSet.modelRelationships) > 0:
                            for table in tblAxisRelSet.rootConcepts:
                                # find a table group object corresponding to this table (ugly hack?)
                                for tableGroupConcept in self.modelXbrl.nameConcepts.get( table.id.replace("_tQ_", "_tgQ_"), () ):
                                    self.tablesToELR[tableGroupConcept.objectId()] = lytMdlTableSet.srcLinkrole
        self.tablesMenu.delete(0, self.tablesMenuLength)
        self.tablesMenuLength = 0
        self.tblELR = None
        for tblMenuEntry in sorted(self.tblMenuEntries.items()):
            tbl,elr = tblMenuEntry
            self.tablesMenu.add_command(label=tbl, command=lambda e=elr: self.view(viewTblELR=e)) # use this to activate profiling from menu selection:  , profile=True))
            self.tablesMenuLength += 1
            if self.tblELR is None:
                self.tblELR = elr # start viewing first ELR

    def viewReloadDueToMenuAction(self, *args):
        if not self.blockMenuEvents:
            # update config (config saved when exiting)
            self.options["ignoreDimValidity"] = self.ignoreDimValidity.get()
            self.view()

    def setOpenBreakdownEntryRows(self, *args):
        import tkinter.simpledialog
        newValue = tkinter.simpledialog.askinteger(_("arelle - Open breakdown entry rows setting"),
                _("The number of extra entry rows for open breakdowns is: {0} \n\n"
                  "(When a row header includes an open breakdown, such as \nfor typed dimension(s), this number of extra entry rows \nare provided below the table.)"
                  ).format(self.options["openBreakdownLines"]),
                parent=self.tabWin)
        if newValue is not None:
            self.options["openBreakdownLines"] = self.openBreakdownLines = newValue
            self.viewReloadDueToMenuAction()

    def view(self, viewTblELR=None, newInstance=None, profile=False):
        startedAt = time.time()
        self.blockMenuEvents += 1
        if newInstance is not None:
            self.modelXbrl = newInstance # a save operation has created a new instance to use subsequently
            clearZchoices = False
        if viewTblELR:  # specific table selection
            self.tblELR = viewTblELR
            clearZchoices = True
        else:   # first or subsequenct reloading (language, dimensions, other change)
            clearZchoices = self.zHdrElts is None

        # remove old widgets
        self.viewFrame.clearGrid()

        layoutTable(self)

        lytMdlTblMdl = self.lytMdlTblMdl
        if len(lytMdlTblMdl.lytMdlTableSets) == 0 or len(lytMdlTblMdl.lytMdlTableSets[0].lytMdlTables) == 0:
            if TRACE_TK: print("no table to display")
            self.blockMenuEvents -= 1
            return # no table to display

        if clearZchoices: # also need first time initialization
            self.loadTablesMenu()  # load menus (and initialize if first time
            self.zHdrElts = None
        # identify tableSet in tableSets
        viewTblELR = viewTblELR or self.tblELR
        for lytMdlTableSet in self.lytMdlTblMdl.lytMdlTableSets:
            if lytMdlTableSet.srcLinkrole == viewTblELR:
                break
        self.lytMdlTable = lytMdlTableSet.lytMdlTables[0] # only one table in table set?
        self.zConstraints = set()
        self.xConstraints = defaultdict(set) # index by xColNum
        self.yConstraints = defaultdict(set) # index by yRowNum
        if self.zHdrElts is None:
            # each Z is a separate table in the outer table
            lytMdlZHdrs = self.lytMdlTable.lytMdlAxisHeaders("z")
            if lytMdlZHdrs is not None:
                lytMdlZHdrGroups = lytMdlZHdrs.lytMdlGroups
                numZtbls = self.lytMdlTable.numBodyCells("z") or 1 # must have at least 1 z entry
                self.zHdrElts = [OrderedDict() for i in range(numZtbls)]
                self.zAspectChoices = OrderedDict()
                for lytMdlZGrp in lytMdlZHdrGroups:
                    for lytMdlZHdr in lytMdlZGrp.lytMdlHeaders:
                        zRow = 0
                        if all(lytMdlCell.isOpenAspectEntrySurrogate for lytMdlCell in lytMdlZHdr.lytMdlCells):
                            continue # skip header with only open aspect entry surrogate
                        for iz, lytMdlZCell in enumerate(lytMdlZHdr.lytMdlCells):
                            zConstraint = tuple(c.aspect for c in lytMdlZCell.lytMdlConstraints)
                            for iSpan in range(lytMdlZCell.span):
                                if not lytMdlZCell.rollup:
                                    zAspLbl = ", ".join(lbl[0] for lbl in lytMdlZCell.labels)
                                    self.zHdrElts[zRow][zConstraint] = zAspLbl
                                    if zConstraint not in self.zAspectChoices:
                                        self.zAspectChoices[zConstraint] = defaultdict(set)
                                    self.zAspectChoices[zConstraint][zAspLbl].add(zRow)
                                zRow += 1
                            self.zConstraints.update(lytMdlZCell.lytMdlConstraints)
            else:
                self.zHdrElts = [[]]
                numZtbls = 1
            self.zTbl = 0
            # get number of y header columns
            self.numZHdrs = self.numYHdrCols = self.numXHdrRows = 0
            self.dataRows = self.dataRows = 0
            lytMdlZHdrs = self.lytMdlTable.lytMdlAxisHeaders("z")
            if lytMdlZHdrs is not None:
                '''
                for lytMdlZGrp in lytMdlZHdrs.lytMdlGroups:
                    for lytMdlZHdr in lytMdlZGrp.lytMdlHeaders:
                        dataRow = 0
                        for lytMdlZCell in lytMdlZHdr.lytMdlCells:
                            dataRow += 1
                        if dataRow > self.numZHdrs:
                            self.numZHdrs = dataRow
                '''
                dataRow = len(lytMdlZHdrs.lytMdlGroups)
                if dataRow > self.numZHdrs:
                    self.numZHdrs = dataRow
            for lytMdlYGrp in getattr(self.lytMdlTable.lytMdlAxisHeaders("y"), "lytMdlGroups", []):
                for lytMdlYHdr in lytMdlYGrp.lytMdlHeaders:
                    self.numYHdrCols +=  lytMdlYHdr.maxNumLabels
                    dataRow = 0
                    if all(lytMdlCell.isOpenAspectEntrySurrogate for lytMdlCell in lytMdlYHdr.lytMdlCells):
                        continue # skip header with only open aspect entry surrogate
                    for lytMdlYCell in lytMdlYHdr.lytMdlCells:
                        dataRow += lytMdlYCell.span
                    if dataRow > self.dataRows:
                        self.dataRows = dataRow
            for lytMdlXGrp in getattr(self.lytMdlTable.lytMdlAxisHeaders("x"), "lytMdlGroups", []):
                for lytMdlXHdr in lytMdlXGrp.lytMdlHeaders:
                    self.numXHdrRows += lytMdlXHdr.maxNumLabels
                    dataCol = 0
                    if all(lytMdlCell.isOpenAspectEntrySurrogate for lytMdlCell in lytMdlXHdr.lytMdlCells):
                        continue # skip header with only open aspect entry surrogate
                    for lytMdlCell in lytMdlXHdr.lytMdlCells:
                        if lytMdlCell.isOpenAspectEntrySurrogate:
                            continue # strip all open aspect entry surrogates from layout model file
                        dataCol += lytMdlCell.span
                    if dataCol > self.dataRows:
                        self.dataRows = dataCol
        self.dataFirstRow = self.numZHdrs + self.numXHdrRows
        if TRACE_TK: print(f"resizeTable rows {self.dataFirstRow+self.dataRows} cols {self.numYHdrCols+self.dataRows} titleRows {self.dataFirstRow} titleColumns {self.numYHdrCols})")
        self.table.resizeTable(self.dataFirstRow+self.dataRows, self.numYHdrCols+self.dataRows, titleRows=self.dataFirstRow, titleColumns=self.numYHdrCols)

        try:
            # review row header wrap widths and limit to 2/3 of the frame width (all are screen units)
            fontWidth = tkFont.Font(font='TkTextFont').configure()['size']
            fontWidth = fontWidth * 3 // 2
            dataColsAllowanceWidth = (fontWidth * ENTRY_WIDTH_IN_CHARS + PADDING) * self.dataCols + PADDING
            frameWidth = self.viewFrame.winfo_width()
            if dataColsAllowanceWidth + self.rowHdrWrapLength > frameWidth:
                if dataColsAllowanceWidth > frameWidth / 2:
                    rowHdrAllowanceWidth = frameWidth / 2
                else:
                    rowHdrAllowanceWidth = frameWidth - dataColsAllowanceWidth
                if self.rowHdrWrapLength > rowHdrAllowanceWidth:
                    widthRatio = rowHdrAllowanceWidth / self.rowHdrWrapLength
                    self.rowHdrWrapLength = rowHdrAllowanceWidth
                    fixedWidth = sum(w for w in self.rowHdrColWidth if w <= RENDER_UNITS_PER_CHAR)
                    adjustableWidth = sum(w for w in self.rowHdrColWidth if w > RENDER_UNITS_PER_CHAR)
                    if adjustableWidth> 0:
                        widthRatio = (rowHdrAllowanceWidth - fixedWidth) / adjustableWidth
                        for i in range(len(self.rowHdrColWidth)):
                            w = self.rowHdrColWidth[i]
                            if w > RENDER_UNITS_PER_CHAR:
                                self.rowHdrColWidth[i] = int(w * widthRatio)
            self.aspectEntryObjectIdsNode.clear()
            self.aspectEntryObjectIdsCell.clear()
            self.factPrototypeAspectEntryObjectIds.clear()
            if TRACE_TK: print(f"tbl hdr x {0} y {0} cols {self.numYHdrCols-1} rows {self.dataFirstRow - 1} value {lytMdlTableSet.label}")
            self.table.initHeaderCellValue(lytMdlTableSet.label,
                                           0, 0, self.numYHdrCols-1, self.dataFirstRow - 1,
                                           XbrlTable.TG_TOP_LEFT_JUSTIFIED)
            self.zAxis(clearZchoices, self.numXHdrRows)
            self.xAxis(self.numYHdrCols, self.colHdrTopRow)
            self.yAxis(0, self.dataFirstRow)
            for fp in self.factPrototypes: # dereference prior facts
                if fp is not None:
                    fp.clear()
            self.factPrototypes = []

            startedAt2 = time.time()
            if self.bodyCells(self.numYHdrCols, self.dataFirstRow):
                # has body cells
                #print("bodyCells {:.2f}secs ".format(time.time() - startedAt2) + self.roledefinition)
                self.table.clearModificationStatus()
                self.table.disableUnusedCells()
                self.table.resizeTableCells()

            # data cells
            #print("body cells done")
        except Exception as err:
            self.modelXbrl.error(f"exception: {type(err).__name__}",
                "Table Linkbase GUI rendering exception: %(error)s at %(traceback)s",
                modelXbrl=self.modelXbrl, error=err,
                traceback=traceback.format_exc())

        self.modelXbrl.profileStat("viewTable_" + os.path.basename(viewTblELR), time.time() - startedAt)
        self.blockMenuEvents -= 1

    def zAxis(self, clearZchoices, colSpan):
        if not self.zHdrElts or not self.zHdrElts[0]:
            return
        # create combo box cells for multiple choice elements
        for iZ, (aspect, aspectChoices) in enumerate(self.zAspectChoices.items()):
            values = [a for a in aspectChoices.keys() if a and a != OPEN_ASPECT_ENTRY_SURROGATE]
            if TRACE_TK: print(f"zAxis comboBox x {self.dataFirstCol} y {iZ} values {values} value {self.zHdrElts[self.zTbl][aspect]} colspan {colSpan}")
            if values:
                combobox = self.table.initHeaderCombobox(self.dataFirstCol,
                                                         iZ,
                                                         values=values,
                                                         value=self.zHdrElts[self.zTbl][aspect],
                                                         colspan=colSpan,
                                                         #selectindex=self.zBreakdownLeafNbr[breakdownRow],
                                                         comboboxselected=self.onZComboBoxSelected)
                combobox.zAspect = aspect

    def onZComboBoxSelected(self, event):
        combobox = event.widget
        comboAspect = combobox.zAspect
        comboVal = combobox.get()
        zTblsForAspect = self.zAspectChoices[combobox.zAspect][comboVal]
        # find zRow for the selected aspects
        for asp, aspVal in self.zHdrElts[self.zTbl].items():
            if asp != comboAspect:
                zTblsForAspect &= self.zAspectChoices[asp][aspVal]
        if len(zTblsForAspect) > 0:
            self.zTbl = next(iter(zTblsForAspect)) # get first of set
            self.view() # redraw grid

    def xAxis(self, leftCol, topRow):
        lytMdlXHdrs = self.lytMdlTable.lytMdlAxisHeaders("x")
        if lytMdlXHdrs is None:
            return
        firstColHdr = True
        yValue = topRow
        for lytMdlGrp in lytMdlXHdrs.lytMdlGroups:
            for iHdr, lytMdlHdr in enumerate(lytMdlGrp.lytMdlHeaders):
                if all(lytMdlCell.isOpenAspectEntrySurrogate for lytMdlCell in lytMdlHdr.lytMdlCells):
                    continue # skip header with only open aspect entry surrogate
                for iLabel in range(lytMdlHdr.maxNumLabels):
                    xValue = leftCol
                    for lytMdlCell in lytMdlHdr.lytMdlCells:
                        if lytMdlCell.isOpenAspectEntrySurrogate:
                            continue # strip all open aspect entry surrogates from layout model file
                        if TRACE_TK: print(f"xAxis hdr x {xValue} y {yValue} cols {lytMdlCell.span} rows {1} isRollUpParent {lytMdlCell.rollup} value \"{lytMdlCell.labelXmlText(iLabel,'')}\"")
                        self.table.initHeaderCellValue(lytMdlCell.labelXmlText(iLabel,""),
                                                       xValue, yValue,
                                                       lytMdlCell.span - 1,
                                                       0, # rowspan - 1,
                                                       XbrlTable.TG_CENTERED,
                                                       #objectId=xStrctNode.objectId(),
                                                       hasTopBorder=not lytMdlCell.rollup,
                                                       hasBottomBorder = (
                                                           iHdr >= len(lytMdlGrp.lytMdlHeaders) - 1 or
                                                           not any(nxtHdrCell.rollup
                                                                   for nxtHdrCell in lytMdlGrp.lytMdlHeaders[iHdr+1].lytMdlCells))
                                                       )
                        self.xConstraints[xValue].update(lytMdlCell.lytMdlConstraints)
                        xValue += lytMdlCell.span or 1
                    yValue += 1

    def yAxis(self, leftCol, row):
        lytMdlYHdrs = self.lytMdlTable.lytMdlAxisHeaders("y")
        if lytMdlYHdrs is None:
            return
        xValue = leftCol
        for lytMdlGrp in lytMdlYHdrs.lytMdlGroups:
            for iHdr, lytMdlHdr in enumerate(lytMdlGrp.lytMdlHeaders):
                yRow = 0
                if all(lytMdlCell.isOpenAspectEntrySurrogate for lytMdlCell in lytMdlHdr.lytMdlCells):
                    continue # skip header with only open aspect entry surrogate
                for lytMdlCell in lytMdlHdr.lytMdlCells:
                    if lytMdlCell.isOpenAspectEntrySurrogate:
                        continue # strip all open aspect entry surrogates from layout model file
                    for iLabel in range(lytMdlHdr.maxNumLabels):
                        if TRACE_TK: print(f"yAxis hdr x {xValue + iLabel} y {row + yRow} cols {1} rows {lytMdlCell.span} rollup {lytMdlCell.rollup} value \"{lytMdlCell.labelXmlText(iLabel,'')}\"")
                        self.table.initHeaderCellValue(lytMdlCell.labelXmlText(iLabel,""),
                                                       xValue + iLabel, row + yRow,
                                                       0, #columnspan + nestedColumnspan - 1,
                                                       lytMdlCell.span - 1, #nestedRowspan - 1,
                                                       XbrlTable.TG_LEFT_JUSTIFIED,
                                                       #objectId=yStrctNode.objectId(),
                                                       hasLeftBorder = not lytMdlCell.rollup,
                                                       hasRightBorder = (
                                                           iHdr >= len(lytMdlGrp.lytMdlHeaders) - 1 or
                                                           not any(nxtHdrCell.rollup
                                                                   for nxtHdrCell in lytMdlGrp.lytMdlHeaders[iHdr+1].lytMdlCells))
                                                       #width=3 if lytMdlCell.rollup else None
                                                       )
                        self.yConstraints[row + yRow].update(lytMdlCell.lytMdlConstraints)
                    yRow += lytMdlCell.span
                xValue += lytMdlHdr.maxNumLabels

    def getbackgroundColor(self, factPrototype):
        bgColor = XbrlTable.TG_BG_DEFAULT # default monetary
        concept = factPrototype.concept
        if concept == None:
            return bgColor
        isNumeric = concept.isNumeric
        # isMonetary = concept.isMonetary
        isInteger = concept.baseXbrliType in integerItemTypes
        isPercent = concept.typeQname in (qnPercentItemType, qnPureItemType)
        isString = concept.baseXbrliType in ("stringItemType", "normalizedStringItemType")
        isDate = concept.baseXbrliType in ("dateTimeItemType", "dateItemType")
        if isNumeric:
            if concept.isShares or isInteger:
                bgColor = XbrlTable.TG_BG_ORANGE
            elif isPercent:
                bgColor = XbrlTable.TG_BG_YELLOW
            # else assume isMonetary
        elif isDate:
            bgColor = XbrlTable.TG_BG_GREEN
        elif isString:
            bgColor = XbrlTable.TG_BG_VIOLET
        return bgColor;

    def bodyCells(self, leftCol, topRow):
        if not self.lytMdlTable.lytMdlBodyChildren:
            return False # no body cells
        lytMdlZBodyCell = self.lytMdlTable.lytMdlBodyChildren[0] # examples only show one z cell despite number of tables
        if self.zTbl >= len(lytMdlZBodyCell.lytMdlBodyChildren):
            return False # no body cells for selected z table
        lytMdlYBodyCell = lytMdlZBodyCell.lytMdlBodyChildren[self.zTbl]
        yRowNum = topRow
        for lytMdlXBodyCell in lytMdlYBodyCell.lytMdlBodyChildren:
            if not any(lytMdlCell.isOpenAspectEntrySurrogate for lytMdlCell in lytMdlXBodyCell.lytMdlBodyChildren):
                for iBodyCell, lytMdlCell in enumerate(lytMdlXBodyCell.lytMdlBodyChildren):
                    xColNum = leftCol + iBodyCell
                    if lytMdlCell.isOpenAspectEntrySurrogate:
                        continue
                    justify = "left"
                    f = None # should be a fact prototype for cell with no bound facts and not blocked
                    for f, v, justify in lytMdlCell.facts:
                        break;
                    #colElt = etree.SubElement(rowElt, "{http://www.w3.org/1999/xhtml}td",
                    #                          attrib={"class":"cell",
                    #                                  "style":f"text-align:{justify};width:8em;"}
                    #                 ).text = "\n".join(v for f, v, justify in lytMdlCell.facts)
                    if f is not None:
                        fp = f
                        value = v
                        objectId = f.objectId()
                    else:
                        cellAspectValues = dict((c.aspect, c.value)
                                                for aC in (self.zConstraints, self.xConstraints[xColNum], self.yConstraints[yRowNum])
                                                for c in aC)
                        fp = FactPrototype(self, cellAspectValues)
                        value = None
                        objectId = "f{0}".format(len(self.factPrototypes))
                        self.factPrototypes.append(fp)  # for property views
                        for aspect, aspectValue in cellAspectValues.items():
                            if isinstance(aspectValue, str) and aspectValue.startswith(OPEN_ASPECT_ENTRY_SURROGATE):
                                self.factPrototypeAspectEntryObjectIds[objectId].add(aspectValue)
                    if fp is not None and fp.concept is not None and not fp.concept.isAbstract:
                        modelConcept = fp.concept
                        if (justify is None) and modelConcept is not None:
                            justify = XbrlTable.TG_RIGHT_JUSTIFIED if modelConcept.isNumeric else XbrlTable.TG_LEFT_JUSTIFIED
                        if modelConcept is not None and modelConcept.isEnumeration:
                            myValidationObject = ValidateXbrl(self.modelXbrl)
                            myValidationObject.modelXbrl = self.modelXbrl
                            enumerationSet = ValidateXbrlDimensions.usableEnumerationMembers(myValidationObject, modelConcept)
                            enumerationDict = dict()
                            for enumerationItem in enumerationSet:
                                # we need to specify the concept linkrole to sort out between possibly many different labels
                                enumerationDict[enumerationItem.label(linkrole=modelConcept.enumLinkrole)] = enumerationItem.qname
                            enumerationValues = sorted(list(enumerationDict.keys()))
                            enumerationQNameStrings = [""]+list(str(enumerationDict[enumerationItem]) for enumerationItem in enumerationValues)
                            enumerationValues = [""]+enumerationValues
                            try:
                                selectedIdx = enumerationQNameStrings.index(value)
                                effectiveValue = enumerationValues[selectedIdx]
                            except ValueError:
                                effectiveValue = enumerationValues[0]
                                selectedIdx = 0
                            if TRACE_TK: print(f"body comboBox enums x {xColNum} y {yRowNum} values {enumerationValues} value {effectiveValue}")
                            self.table.initCellCombobox(effectiveValue,
                                                        enumerationValues,
                                                        xColNum,
                                                        yRowNum,
                                                        objectId=objectId,
                                                        selectindex=selectedIdx,
                                                        codes=enumerationDict)
                        elif modelConcept is not None and modelConcept.type is not None and modelConcept.type.qname == XbrlConst.qnXbrliQNameItemType:
                            if eurofilingModelPrefix in modelConcept.nsmap and modelConcept.nsmap.get(eurofilingModelPrefix) == eurofilingModelNamespace:
                                hierarchy = modelConcept.get("{" + eurofilingModelNamespace + "}" + "hierarchy", None)
                                domainQNameAsString = modelConcept.get("{" + eurofilingModelNamespace + "}" + "domain", None)
                                if hierarchy is not None and domainQNameAsString is not None:
                                    newAspectValues = [""]
                                    newAspectQNames = dict()
                                    newAspectQNames[""] = None
                                    domPrefix, _, domLocalName = domainQNameAsString.strip().rpartition(":")
                                    domNamespace = modelConcept.nsmap.get(domPrefix)
                                    relationships = concept_relationships(self.rendrCntx,
                                         None,
                                         (QName(domPrefix, domNamespace, domLocalName),
                                          hierarchy, # linkrole,
                                          "XBRL-dimensions",
                                          'descendant'),
                                         False) # return flat list
                                    for rel in relationships:
                                        if (rel.arcrole in (XbrlConst.dimensionDomain, XbrlConst.domainMember)
                                            and rel.isUsable):
                                            header = rel.toModelObject.label(lang=self.lang)
                                            newAspectValues.append(header)
                                            currentQName = rel.toModelObject.qname
                                            if str(currentQName) == value:
                                                value = header
                                            newAspectQNames[header] = currentQName
                                else:
                                    newAspectValues = None
                            else:
                                newAspectValues = None
                            if newAspectValues is None:
                                self.table.initCellValue(value,
                                                         xColNum,
                                                         yRowNum,
                                                         justification=justify,
                                                         objectId=objectId,
                                                         backgroundColourTag=self.getbackgroundColor(fp))
                            else:
                                qNameValues = newAspectValues
                                try:
                                    selectedIdx = qNameValues.index(value)
                                    effectiveValue = value
                                except ValueError:
                                    effectiveValue = qNameValues[0]
                                    selectedIdx = 0
                                if TRACE_TK: print(f"body comboBox qnames x {xColNum} y {yRowNum} values {qNameValues} value {effectiveValue}")
                                self.table.initCellCombobox(effectiveValue,
                                                            qNameValues,
                                                            xColNum,
                                                            yRowNum,
                                                            objectId=objectId,
                                                            selectindex=selectedIdx,
                                                            codes=newAspectQNames)
                        elif modelConcept is not None and modelConcept.type is not None and modelConcept.type.qname == XbrlConst.qnXbrliBooleanItemType:
                            booleanValues = ["",
                                             XbrlConst.booleanValueTrue,
                                             XbrlConst.booleanValueFalse]
                            try:
                                selectedIdx = booleanValues.index(value)
                                effectiveValue = value
                            except ValueError:
                                effectiveValue = booleanValues[0]
                                selectedIdx = 0
                            if TRACE_TK: print(f"body comboBox bools x {xColNum} y {yRowNum} values {booleanValues} value {effectiveValue}")
                            self.table.initCellCombobox(effectiveValue,
                                                        booleanValues,
                                                        xColNum,
                                                        yRowNum,
                                                        objectId=objectId,
                                                        selectindex=selectedIdx)
                        else:
                            if TRACE_TK: print(f"body cell x {xColNum} y {yRowNum} value {value}")
                            self.table.initCellValue(value,
                                                     xColNum,
                                                     yRowNum,
                                                     justification=justify,
                                                     objectId=objectId,
                                                     backgroundColourTag=self.getbackgroundColor(fp))
                yRowNum += 1
        return True # has body cells

    def onClick(self, event):
        try:
            objId = event.widget.objectId
            if objId and objId[0] == "f":
                viewableObject = self.factPrototypes[int(objId[1:])]
            else:
                viewableObject = objId
            self.modelXbrl.viewModelObject(viewableObject)
        except AttributeError: # not clickable
            pass
        self.modelXbrl.modelManager.cntlr.currentView = self

    def cellEnter(self, *args):
        # triggered on grid frame enter (not cell enter)
        self.blockSelectEvent = 0
        self.modelXbrl.modelManager.cntlr.currentView = self

    def cellLeave(self, *args):
        # triggered on grid frame leave (not cell leave)
        self.blockSelectEvent = 1

    # this method is not currently used
    def cellSelect(self, *args):
        if self.blockSelectEvent == 0 and self.blockViewModelObject == 0:
            self.blockViewModelObject += 1
            #self.modelXbrl.viewModelObject(self.nodeToObjectId[self.treeView.selection()[0]])
            #self.modelXbrl.viewModelObject(self.treeView.selection()[0])
            self.blockViewModelObject -= 1

    def viewModelObject(self, modelObject):
        if self.blockViewModelObject == 0:
            self.blockViewModelObject += 1
            try:
                if isinstance(modelObject, ModelDtsObject.ModelRelationship):
                    objectId = modelObject.toModelObject.objectId()
                elif isinstance(modelObject, ModelDtsObject.ModelRoleType):
                    objectId = self.modelXbrl.roleTypeDefinition(modelObject.roleURI, self.lang)
                else:
                    objectId = modelObject.objectId()
                if objectId in self.tablesToELR:
                    self.view(viewTblELR=self.tablesToELR[objectId])
                    try:
                        self.modelXbrl.modelManager.cntlr.currentView = self.modelXbrl.guiViews.tableView
                        # force focus (synch) on the corresponding "Table" tab (useful in case of several instances)
                        self.modelXbrl.guiViews.tableView.tabWin.select(str(self.modelXbrl.guiViews.tableView.viewFrame))
                    except:
                        pass
            except (KeyError, AttributeError):
                    pass
            self.blockViewModelObject -= 1

    def onConfigure(self, event, *args):
        if not self.blockMenuEvents:
            lastFrameWidth = getattr(self, "lastFrameWidth", 0)
            lastFrameHeight = getattr(self, "lastFrameHeight", 0)
            frameWidth = self.tabWin.winfo_width()
            frameHeight = self.tabWin.winfo_height()
            if lastFrameWidth != frameWidth or lastFrameHeight != frameHeight:
                self.updateInstanceFromFactPrototypes()
                self.lastFrameWidth = frameWidth
                self.lastFrameHeight = frameHeight
                self.setHeightAndWidth()
                if lastFrameWidth:
                    # frame resized, recompute row header column widths and lay out table columns
                    """
                    def sleepAndReload():
                        time.sleep(.75)
                        self.viewReloadDueToMenuAction()
                    self.modelXbrl.modelManager.cntlr.uiThreadQueue.put((sleepAndReload, []))
                    """
                    #self.modelXbrl.modelManager.cntlr.uiThreadQueue.put((self.viewReloadDueToMenuAction, []))
                    def deferredReload():
                        self.deferredReloadCount -= 1  # only do reload after all queued reload timers expire
                        if self.deferredReloadCount <= 0:
                            self.viewReloadDueToMenuAction()
                    self.deferredReloadCount = getattr(self, "deferredReloadCount", 0) + 1
                    self.viewFrame.after(1500, deferredReload)

    def onQuitView(self, event, *args):
        # this method is passed as callback when creating the view
        # (to ScrolledTkTableFrame and then to XbrlTable that will monitor cell operations)
        self.updateInstanceFromFactPrototypes()
        self.updateProperties()

    def hasChangesToSave(self):
        return len(self.table.modifiedCells)

    def updateProperties(self):
        if self.modelXbrl is not None:
            modelXbrl =  self.modelXbrl
            # make sure we handle an instance
            if modelXbrl.modelDocument.type == ModelDocumentType.INSTANCE:
                tbl = self.table
                # get coordinates of last currently operated cell
                coordinates = tbl.getCurrentCellCoordinates()
                if coordinates is not None:
                    # get object identifier from its coordinates in the current table
                    objId = tbl.getObjectId(coordinates)
                    if objId is not None and len(objId) > 0:
                        if objId and objId[0] == "f":
                            # fact prototype
                            viewableObject = self.factPrototypes[int(objId[1:])]
                        elif objId[0] != "a":
                            # instance fact
                            viewableObject = self.modelXbrl.modelObject(objId)
                        else:
                            return
                        modelXbrl.viewModelObject(viewableObject)


    def updateInstanceFromFactPrototypes(self):
        # Only update the model if it already exists
        if self.modelXbrl is not None \
           and self.modelXbrl.modelDocument.type == ModelDocumentType.INSTANCE:
            instance = self.modelXbrl
            cntlr =  instance.modelManager.cntlr
            newCntx = ModelXbrl.AUTO_LOCATE_ELEMENT
            newUnit = ModelXbrl.AUTO_LOCATE_ELEMENT
            tbl = self.table
            # check user keyed changes to aspects
            aspectEntryChanges = {}  # index = widget ID,  value = widget contents
            aspectEntryChangeIds = aspectEntryChanges.keys()
            for modifiedCell in tbl.getCoordinatesOfModifiedCells():
                objId = tbl.getObjectId(modifiedCell)
                if objId is not None and len(objId)>0:
                    if tbl.isHeaderCell(modifiedCell):
                        if objId[0] == OPEN_ASPECT_ENTRY_SURROGATE:
                            aspectEntryChanges[objId] = tbl.getTableValue(modifiedCell)
                    else:
                        # check user keyed changes to facts
                        cellIndex = str(modifiedCell)
                        comboboxCells = tbl.window_names(cellIndex)
                        if comboboxCells is not None and len(comboboxCells)>0:
                            comboName = tbl.window_cget(cellIndex, '-window')
                            combobox = cntlr.parent.nametowidget(comboName)
                        else:
                            combobox = None
                        if isinstance(combobox, _Combobox):
                            codeDict = combobox.codes
                            if len(codeDict)>0: # the drop-down list shows labels, we want to have the actual values
                                bodyCellValue = tbl.getTableValue(modifiedCell)
                                value = codeDict.get(bodyCellValue, None)
                                if value is None:
                                    value = bodyCellValue # this must be a qname!
                            else:
                                value = tbl.getTableValue(modifiedCell)
                        else:
                            value = tbl.getTableValue(modifiedCell)
                        objId = tbl.getObjectId(modifiedCell)
                        if objId is not None and len(objId)>0:
                            if objId[0] == "f":
                                factPrototypeIndex = int(objId[1:])
                                factPrototype = self.factPrototypes[factPrototypeIndex]
                                concept = factPrototype.concept
                                if concept is None:
                                    if not self.conceptMessageIssued:
                                        # This should be removed once cells have been disabled until every needed selection is done
                                        self.conceptMessageIssued = True
                                        self.modelXbrl.modelManager.cntlr.showMessage(_("Please make sure every Z axis selection is done"))
                                    return
                                else:
                                    self.conceptMessageIssued = False
                                entityIdentScheme = self.newFactItemOptions.entityIdentScheme
                                entityIdentValue = self.newFactItemOptions.entityIdentValue
                                periodType = concept.periodType
                                periodStart = self.newFactItemOptions.startDateDate if periodType == "duration" else None
                                periodEndInstant = self.newFactItemOptions.endDateDate
                                qnameDims = factPrototype.context.qnameDims
                                newAspectValues = self.newFactOpenAspects(objId)
                                if newAspectValues is None:
                                    self.modelXbrl.modelManager.showStatus(_("Some open values are missing in an axis, the save is incomplete"), 5000)
                                    continue
                                qnameDims.update(newAspectValues)
                                # open aspects widgets
                                prevCntx = instance.matchContext(
                                    entityIdentScheme, entityIdentValue, periodType, periodStart, periodEndInstant,
                                    qnameDims, [], [])
                                if prevCntx is not None:
                                    cntxId = prevCntx.id
                                else: # need new context
                                    newCntx = instance.createContext(entityIdentScheme, entityIdentValue,
                                        periodType, periodStart, periodEndInstant,
                                        concept.qname, qnameDims, [], [],
                                        afterSibling=newCntx)
                                    cntxId = newCntx.id # need new context
                                # new context
                                if concept.isNumeric:
                                    if concept.isMonetary:
                                        unitMeasure = qname(XbrlConst.iso4217, self.newFactItemOptions.monetaryUnit)
                                        unitMeasure.prefix = "iso4217" # want to save with a recommended prefix
                                        decimals = self.newFactItemOptions.monetaryDecimals
                                    elif concept.isShares:
                                        unitMeasure = XbrlConst.qnXbrliShares
                                        decimals = self.newFactItemOptions.nonMonetaryDecimals
                                    else:
                                        unitMeasure = XbrlConst.qnXbrliPure
                                        decimals = self.newFactItemOptions.nonMonetaryDecimals
                                    prevUnit = instance.matchUnit([unitMeasure], [])
                                    if prevUnit is not None:
                                        unitId = prevUnit.id
                                    else:
                                        newUnit = instance.createUnit([unitMeasure], [], afterSibling=newUnit)
                                        unitId = newUnit.id
                                attrs = [("contextRef", cntxId)]
                                if concept.isNumeric:
                                    attrs.append(("unitRef", unitId))
                                    attrs.append(("decimals", decimals))
                                    value = Locale.atof(self.modelXbrl.locale, value, str.strip)
                                newFact = instance.createFact(concept.qname, attributes=attrs, text=value)
                                tbl.setObjectId(modifiedCell,
                                                newFact.objectId()) # switch cell to now use fact ID
                                if self.factPrototypes[factPrototypeIndex] is not None:
                                    self.factPrototypes[factPrototypeIndex].clear()
                                self.factPrototypes[factPrototypeIndex] = None #dereference fact prototype
                            elif objId[0] != "a": # instance fact, not prototype
                                fact = self.modelXbrl.modelObject(objId)
                                if isinstance(fact, ModelInstanceObject.ModelFact):
                                    if fact.concept.isNumeric:
                                        value = Locale.atof(self.modelXbrl.locale, value, str.strip)
                                        if fact.concept.isMonetary:
                                            unitMeasure = qname(XbrlConst.iso4217, self.newFactItemOptions.monetaryUnit)
                                            unitMeasure.prefix = "iso4217" # want to save with a recommended prefix
                                            decimals = self.newFactItemOptions.monetaryDecimals
                                        elif fact.concept.isShares:
                                            unitMeasure = XbrlConst.qnXbrliShares
                                            decimals = self.newFactItemOptions.nonMonetaryDecimals
                                        else:
                                            unitMeasure = XbrlConst.qnXbrliPure
                                            decimals = self.newFactItemOptions.nonMonetaryDecimals
                                    if fact.value != str(value):
                                        if fact.isNil != (not value):
                                            fact.isNil = not value
                                            if fact.isNil:
                                                pass
                                                #TODO: clear out nil facts
                                        if fact.concept.isNumeric and (not fact.isNil): # if nil, there is no need to update these values
                                            fact.decimals = decimals
                                            prevUnit = instance.matchUnit([unitMeasure], [])
                                            if prevUnit is not None:
                                                unitId = prevUnit.id
                                            else:
                                                newUnit = instance.createUnit([unitMeasure], [], afterSibling=newUnit)
                                                unitId = newUnit.id
                                            fact.unitID = unitId
                                        fact.text = str(value)
                                        instance.setIsModified()
                                        fact.xValid = UNVALIDATED
                                        xmlValidate(instance, fact)
            tbl.clearModificationStatus()

    def saveInstance(self, newFilename=None, onSaved=None):
        if (not self.newFactItemOptions.entityIdentScheme or  # not initialized yet
            not self.newFactItemOptions.entityIdentValue or
            not self.newFactItemOptions.startDateDate or not self.newFactItemOptions.endDateDate):
            if not getNewFactItemOptions(self.modelXbrl.modelManager.cntlr, self.newFactItemOptions):
                return # new instance not set
        # newFilename = None # only used when a new instance must be created

        self.updateInstanceFromFactPrototypes()
        if self.modelXbrl.modelDocument.type != ModelDocumentType.INSTANCE and newFilename is None:
            newFilename = self.modelXbrl.modelManager.cntlr.fileSave(view=self, fileType="xbrl")
            if not newFilename:
                return  # saving cancelled
        # continue saving in background
        thread = threading.Thread(target=lambda: self.backgroundSaveInstance(newFilename, onSaved))
        thread.daemon = True
        thread.start()


    def backgroundSaveInstance(self, newFilename=None, onSaved=None):
        cntlr = self.modelXbrl.modelManager.cntlr
        if newFilename and self.modelXbrl.modelDocument.type != ModelDocumentType.INSTANCE:
            self.modelXbrl.modelManager.showStatus(_("creating new instance {0}").format(os.path.basename(newFilename)))
            self.modelXbrl.modelManager.cntlr.waitForUiThreadQueue() # force status update
            self.modelXbrl.createInstance(newFilename) # creates an instance as this modelXbrl's entrypoint
        instance = self.modelXbrl
        cntlr.showStatus(_("Saving {0}").format(instance.modelDocument.basename))
        cntlr.waitForUiThreadQueue() # force status update

        self.updateInstanceFromFactPrototypes()
        instance.saveInstance(overrideFilepath=newFilename) # may override prior filename for instance from main menu
        cntlr.addToLog(_("{0} saved").format(newFilename if newFilename is not None else instance.modelDocument.filepath))
        cntlr.showStatus(_("Saved {0}").format(instance.modelDocument.basename), clearAfter=3000)
        if onSaved is not None:
            self.modelXbrl.modelManager.cntlr.uiThreadQueue.put((onSaved, []))

    def newFactOpenAspects(self, factObjectId):
        aspectValues = {}
        for aspectObjId in self.factPrototypeAspectEntryObjectIds[factObjectId]:
            structuralNode = self.aspectEntryObjectIdsNode[aspectObjId]
            for aspect in structuralNode.aspectsCovered():
                if aspect != Aspect.DIMENSIONS:
                    break
            gridCellItem = self.aspectEntryObjectIdsCell[aspectObjId]
            value = gridCellItem.get()
            # is aspect in a childStrctNode?
            if value is not None and OPEN_ASPECT_ENTRY_SURROGATE in aspectObjId and len(value)==0:
                return None # some values are missing!
            if value:
                aspectValue = structuralNode.aspectEntryHeaderValues.get(value)
                if aspectValue is None: # try converting value
                    if isinstance(aspect, QName): # dimension
                        dimConcept = self.modelXbrl.qnameConcepts[aspect]
                        if dimConcept.isExplicitDimension:
                            # value must be qname
                            aspectValue = None # need to find member for the description
                        else:
                            typedDimElement = dimConcept.typedDomainElement
                            aspectValue = FunctionXfi.create_element(
                                  self.rendrCntx, None, (typedDimElement.qname, (), value))
                if aspectValue is not None:
                    aspectValues[aspect] = aspectValue
        return aspectValues

    def aspectEntryValues(self, structuralNode):
        for aspect in structuralNode.aspectsCovered():
            if aspect != Aspect.DIMENSIONS:
                break
        # if findHeader is None, return all header values in a list
        # otherwise return aspect value matching header if any
        depth = 0
        n = structuralNode
        while (n.strctMdlParentNode is not None):
            depth += 1
            root = n = n.strctMdlParentNode

        headers = set()
        headerValues = {}
        def getHeaders(n, d):
            for childStrctNode in n.strctMdlChildNodes:
                if d == depth:
                    h = childStrctNode.header(lang=self.lang,
                                                   returnGenLabel=False,
                                                   returnMsgFormatString=False)
                    if not childStrctNode.isEntryPrototype() and h:
                        headerValues[h] = childStrctNode.aspectValue(aspect)
                        headers.add(h)
                else:
                    getHeaders(childStrctNode, d+1)
        getHeaders(root, 1)

        structuralNode.aspectEntryHeaderValues = headerValues
        # is this an explicit dimension, if so add "(all members)" option at end
        headersList = sorted(headers)
        if isinstance(aspect, QName): # dimension
            dimConcept = self.modelXbrl.qnameConcepts[aspect]
            if dimConcept.isExplicitDimension:
                if headersList: # has entries, add all-memembers at end
                    headersList.append("(all members)")
                else:  # empty list, just add all members anyway
                    return self.explicitDimensionFilterMembers(structuralNode, structuralNode)
        return headersList

    def onAspectComboboxSelection(self, event):
        gridCombobox = event.widget
        if gridCombobox.get() == "(all members)":
            structuralNode = self.aspectEntryObjectIdsNode[gridCombobox.objectId]
            self.comboboxLoadExplicitDimension(gridCombobox, structuralNode, structuralNode)

    def comboboxLoadExplicitDimension(self, gridCombobox, structuralNode, structuralNodeWithFilter):
        gridCombobox["values"] = self.explicitDimensionFilterMembers(structuralNode, structuralNodeWithFilter)

    def explicitDimensionFilterMembers(self, structuralNode, structuralNodeWithFilter):
        for aspect in structuralNodeWithFilter.aspectsCovered():
            if isinstance(aspect, QName): # dimension
                break
        valueHeaders = set()
        if structuralNode is not None:
            headerValues = {}
            # check for dimension filter(s)
            dimFilterRels = structuralNodeWithFilter.defnMdlNode.filterRelationships
            if dimFilterRels:
                for rel in dimFilterRels:
                    dimFilter = rel.toModelObject
                    if dimFilter is not None:
                        for memberModel in dimFilter.memberProgs:
                                memQname = memberModel.qname
                                memConcept = self.modelXbrl.qnameConcepts.get(memQname)
                                if memConcept is not None and (not memberModel.axis or memberModel.axis.endswith('-self')):
                                    header = memConcept.label(lang=self.lang)
                                    valueHeaders.add(header)
                                    if rel.isUsable:
                                        headerValues[header] = memQname
                                    else:
                                        headerValues[header] = memConcept
                                if memberModel.axis and memberModel.linkrole and memberModel.arcrole:
                                    # merge of pull request 42 acsone:TABLE_Z_AXIS_DESCENDANT_OR_SELF
                                    if memberModel.axis.endswith('-or-self'):
                                        searchAxis = memberModel.axis[:len(memberModel.axis)-len('-or-self')]
                                    else:
                                        searchAxis = memberModel.axis
                                    relationships = concept_relationships(self.rendrCntx,
                                                         None,
                                                         (memQname,
                                                          memberModel.linkrole,
                                                          memberModel.arcrole,
                                                          searchAxis),
                                                         False) # return flat list
                                    for rel in relationships:
                                        if rel.isUsable:
                                            header = rel.toModelObject.label(lang=self.lang)
                                            valueHeaders.add(header)
                                            headerValues[header] = rel.toModelObject.qname
            if not valueHeaders:
                relationships = concept_relationships(self.rendrCntx,
                                     None,
                                     (aspect,
                                      "XBRL-all-linkroles", # linkrole,
                                      "XBRL-dimensions",
                                      'descendant'),
                                     False) # return flat list
                for rel in relationships:
                    if (rel.arcrole in (XbrlConst.dimensionDomain, XbrlConst.domainMember)
                        and rel.isUsable):
                        header = rel.toModelObject.label(lang=self.lang)
                        valueHeaders.add(header)
                        headerValues[header] = rel.toModelObject.qname
            structuralNode.aspectEntryHeaderValues = headerValues
        return sorted(valueHeaders)

# import after other modules resolved to prevent circular references
from arelle.FunctionXfi import concept_relationships



"""
See COPYRIGHT.md for copyright information.
"""

from typing import TYPE_CHECKING, Set

from arelle.ModelValue import QName
from arelle.PythonUtil import OrderedSet
from .XbrlProperty import XbrlProperty
from .XbrlTypes import XbrlTaxonomyType, QNameKeyType
from .XbrlTaxonomyObject import XbrlReferencableTaxonomyObject

class XbrlAbstract(XbrlReferencableTaxonomyObject):
    taxonomy: XbrlTaxonomyType
    name: QNameKeyType # (required) The name is a QName that uniquely identifies the abstract object.
    properties: OrderedSet[XbrlProperty] # (optional) ordered set of property objects used to specify additional properties associated with the concept using the property object. Only immutable properties as defined in the propertyType object can be added to a concept.



"""
See COPYRIGHT.md for copyright information.
"""

from typing import TYPE_CHECKING, Optional

from arelle.ModelValue import QName, AnyURI
from arelle.PythonUtil import OrderedSet
from .XbrlProperty import XbrlProperty
from .XbrlTypes import XbrlTaxonomyType, QNameKeyType
from .XbrlTaxonomyObject import XbrlTaxonomyObject, XbrlReferencableTaxonomyObject

class XbrlGroup(XbrlReferencableTaxonomyObject):
    taxonomy: XbrlTaxonomyType
    name: QNameKeyType # (required) The group object QName that identifies the group so it can be referenced by other objects.
    groupURI:Optional[AnyURI] # (optional) The group URI that uniquely identifies the group and id used for backward compatibility with roles.
    properties: OrderedSet[XbrlProperty] # (optional) an ordered set of property objects used to specify additional properties associated with the group using the property object. Only immutable properties as defined in the propertyType object can be added to a group.

class XbrlGroupContent(XbrlTaxonomyObject):
    taxonomy: XbrlTaxonomyType
    groupName: QName # (required) The QName of the group object.
    relatedNames: OrderedSet[QName] # (required) An ordered set of network object or cube object QNames that are included in the group object. The order of the set determines the order they appear in the group. The set cannot be empty. The set can only include the QNames of network and cube objects.



"""
See COPYRIGHT.md for copyright information.
"""

from __future__ import annotations

from collections.abc import Generator, Iterable

from lxml.etree import _Element

from arelle import XbrlConst
from arelle.ModelDtsObject import ModelConcept
from arelle.ModelObject import ModelObject
from arelle.ModelValue import QName
from arelle.ModelXbrl import ModelXbrl
from arelle.typing import TypeGetText

_: TypeGetText


def etreeIterWithDepth(
        node: ModelObject | _Element,
        depth: int = 0,
) -> Generator[tuple[ModelObject | _Element, int], None, None]:
    yield node, depth
    for child in node.iterchildren():
        yield from etreeIterWithDepth(child, depth + 1)


def hasPresentationalConceptsWithFacts(
        modelXbrl: ModelXbrl,
        roleUris: Iterable[str],
        memberQnameFilter: set[QName] | None = None,
) -> bool:
    """
    Returns True if any concepts used in the presentation network of the role URIs have been tagged with facts.
    This DOES NOT check if the facts are dimensionally valid against hypercubes defined in the roles.
    """
    roleRelSet = modelXbrl.relationshipSet(XbrlConst.parentChild, tuple(roleUris))
    concepts = set(roleRelSet.fromModelObjects().keys()) | set(roleRelSet.toModelObjects().keys())
    for concept in concepts:
        if not isinstance(concept, ModelConcept):
            continue
        if concept.qname is None:
            continue
        if concept.isAbstract:
            continue
        for fact in modelXbrl.factsByQname.get(concept.qname, set()):
            if memberQnameFilter is None:
                return True
            if fact.context is not None:
                for dimValue in fact.context.qnameDims.values():
                    if dimValue.memberQname in memberQnameFilter:
                        return True
    return False



"""
See COPYRIGHT.md for copyright information.

Filer Guidelines:
Basically EFM
- 5.2.2.1
- 5.2.2.6
- 5.2.2.10
- 5.2.5.4
- 5.2.5.5 (Inline XBRL)
- 5.2.5.6 (Inline XBRL)
- 5.2.5.7
- 5.2.5.8
- 5.2.5.9
- 5.2.5.10
- 6.5.3
- 6.5.4
- 6.5.7
- 6.5.8
- 6.5.14
- 6.5.15
- 6.5.16
- 6.5.17

Filers can only submit an instance so EFM DTS checks are not needed.
"""
import os
import regex as re
from arelle import ValidateFilingText, XmlUtil
from arelle.ModelDocumentType import ModelDocumentType
from arelle.ModelInstanceObject import ModelFact, ModelInlineFact, ModelInlineFootnote
from arelle.ModelObject import ModelObject
from arelle.PrototypeDtsObject import LinkPrototype, LocPrototype, ArcPrototype
from arelle.Version import authorLabel, copyrightLabel
from arelle.XbrlConst import xbrli, xhtml
from arelle.XmlValidateConst import VALID
from arelle.utils.Contexts import getDuplicateContextGroups

def disclosureSystemTypes(disclosureSystem, *args, **kwargs):
    # return ((disclosure system name, variable name), ...)
    return (("FERC", "FERCplugin"),) # FERC disclosure system

def disclosureSystemConfigURL(disclosureSystem, *args, **kwargs):
    return os.path.join(os.path.dirname(__file__), "config.xml")

def validateXbrlStart(val, parameters=None, *args, **kwargs):
    val.validateFERCplugin = val.validateDisclosureSystem and getattr(val.disclosureSystem, "FERCplugin", False)
    if not (val.validateFERCplugin):
        return

    # use UTR validation if list of URLs was provided
    val.validateUTR = bool(val.disclosureSystem.utrUrl)

def validateXbrlFinally(val, *args, **kwargs):
    if not (val.validateFERCplugin):
        return

    _xhtmlNs = "{{{}}}".format(xhtml)
    _xhtmlNsLen = len(_xhtmlNs)
    modelXbrl = val.modelXbrl
    modelDocument = modelXbrl.modelDocument
    if not modelDocument:
        return # never loaded properly
    disclosureSystem = val.disclosureSystem

    _statusMsg = _("validating {0} filing rules").format(val.disclosureSystem.name)
    modelXbrl.profileActivity()
    modelXbrl.modelManager.showStatus(_statusMsg)

    isInlineXbrl = modelXbrl.modelDocument.type in (ModelDocumentType.INLINEXBRL, ModelDocumentType.INLINEXBRLDOCUMENTSET)
    requiredFactLang = disclosureSystem.defaultXmlLang.lower() if disclosureSystem.defaultXmlLang else disclosureSystem.defaultXmlLang

    # inline doc set has multiple instance names to check
    if modelXbrl.modelDocument.type == ModelDocumentType.INLINEXBRLDOCUMENTSET:
        instanceNames = [ixDoc.basename
                         for ixDoc in modelXbrl.modelDocument.referencesDocument.keys()
                         if ixDoc.type == ModelDocumentType.INLINEXBRL]
        xbrlInstRoots = modelXbrl.ixdsHtmlElements
    else: # single instance document to check is the entry point document
        instanceNames = [modelXbrl.modelDocument.basename]
        xbrlInstRoots = [modelXbrl.modelDocument.xmlDocument.getroot()]

    #6.5.15 facts with xml in text blocks
    ValidateFilingText.validateTextBlockFacts(modelXbrl)

    # check footnotes text
    if isInlineXbrl:
        _linkEltIter = (linkPrototype
                        for linkKey, links in modelXbrl.baseSets.items()
                        for linkPrototype in links
                        if linkPrototype.modelDocument.type in (ModelDocumentType.INLINEXBRL, ModelDocumentType.INLINEXBRLDOCUMENTSET)
                        and linkKey[1] and linkKey[2] and linkKey[3]  # fully specified roles
                        and linkKey[0] != "XBRL-footnotes")
    else:
        _linkEltIter = xbrlInstRoots[0].iterdescendants(tag="{http://www.xbrl.org/2003/linkbase}footnoteLink")
    for footnoteLinkElt in _linkEltIter:
        if isinstance(footnoteLinkElt, (ModelObject,LinkPrototype)):
            for child in footnoteLinkElt:
                if isinstance(child,(ModelObject,LocPrototype,ArcPrototype)):
                    xlinkType = child.get("{http://www.w3.org/1999/xlink}type")
                    if xlinkType == "resource" or isinstance(child,ModelInlineFootnote): # footnote
                        if not isInlineXbrl: # inline content was validated before and needs continuations assembly
                            ValidateFilingText.validateFootnote(modelXbrl, child)

    # same identifier in all contexts (EFM 6.5.3)
    entityIdentifiers = set()
    for xbrlInstRoot in xbrlInstRoots: # check all inline docs in ix doc set
        for entityIdentifierElt in xbrlInstRoot.iterdescendants("{http://www.xbrl.org/2003/instance}identifier"):
            if isinstance(entityIdentifierElt,ModelObject):
                entityIdentifiers.add("{}#{}".format(entityIdentifierElt.get("scheme"), XmlUtil.text(entityIdentifierElt)))
    if len(entityIdentifiers) > 1:
        modelXbrl.error("FERC.6.05.03",
            _("There are more than one entity identifiers: %(entityIdentifiers)s."),
            modelObject=modelXbrl,
            entityIdentifiers=", ".join(sorted(entityIdentifiers)))
    for ei in sorted(entityIdentifiers):
        scheme, _sep, identifier = ei.rpartition("#")
        if not disclosureSystem.identifierSchemePattern.match(scheme) or not disclosureSystem.identifierValuePattern.match(identifier):
            modelXbrl.error("FERC.6.05.01",
                _("Entity identifier %(identifier)s, or scheme %(scheme)s does not adhere "
                  "to the standard naming convention of <identifier scheme='http://www.ferc.gov/CID'>Cnnnnnn</identifier>'.  "),
                modelObject=modelXbrl, scheme=scheme, identifier=identifier)

    #6.5.4 scenario
    segContexts = set()
    contextIDs = set()
    precisionFacts = set()
    formType = None
    formEntrySchema = None
    factsForLang = {}
    keysNotDefaultLang = {}
    allFormEntryXsd = ()

    for c in modelXbrl.contexts.values():
        if XmlUtil.hasChild(c, xbrli, "segment"):
            segContexts.add(c)
        contextIDs.add(c.id)
    for contexts in getDuplicateContextGroups(modelXbrl):
        modelXbrl.error("FERC.6.05.07",
            _("The instance document contained more than one context equivalent to %(context)s (%(context2)s).  "
                "Please remove duplicate contexts from the instance."),
            modelObject=contexts, context=contexts[0].id, context2=contexts[1].id)

    if segContexts:
        modelXbrl.error("FERC.6.05.04",
            _("There must be no contexts with segment, but %(count)s was(were) found: %(context)s."),
            modelObject=segContexts, count=len(segContexts), context=", ".join(sorted(c.id for c in segContexts)))

    # unused contexts
    for f in modelXbrl.facts:
        factContextID = f.contextID
        contextIDs.discard(factContextID)
        if f.isNumeric:
            if f.precision is not None:
                precisionFacts.add(f)
        elif not f.isNil:
            langTestKey = "{0},{1}".format(f.qname, f.contextID)
            factsForLang.setdefault(langTestKey, []).append(f)
            lang = f.xmlLang
            if lang and lang.lower() != requiredFactLang: # not lang.startswith(factLangStartsWith):
                keysNotDefaultLang[langTestKey] = f
        if getattr(f, "xValid", 0) >= VALID:
            if f.qname.localName == "FormType":
                formType = f.xValue
                formNum = re.sub("([0-9]+).*", r"\1", formType)
                formLtr = re.match("[^A-Z]*([A-Z]?)", formType).group(1)
                txDate = re.sub("http://ferc.gov/form/([0-9]{4}-[0-9]{2}-[0-9]{2})/ferc", r"\1", f.qname.namespaceURI)

                formEntryXsd = "https://eCollection.ferc.gov/taxonomy/form{}/{}/form/form{}{}/form-{}{}_{}.xsd".format(formNum, txDate, formNum, formLtr, formNum, formLtr, txDate)

                formEntryXsdUAT = formEntryXsd.replace("eCollection","uat.eforms")
                formEntryXsdTest = formEntryXsd.replace("eCollection","test.eforms")
                formEntryXsdDev = formEntryXsd.replace("eCollection","dev.eforms")

                confFormEntryXsd = "https://eCollection.ferc.gov/taxonomy/form{}/{}/ferc-core-footnote-roles_{}.xsd".format(formNum, txDate,txDate)

                confFormEntryXsdUAT = confFormEntryXsd.replace("eCollection","uat.eforms")
                confFormEntryXsdTest = confFormEntryXsd.replace("eCollection","test.eforms")
                confFormEntryXsdDev = confFormEntryXsd.replace("eCollection","dev.eforms")

                allFormEntryXsd = [formEntryXsd, formEntryXsdUAT, formEntryXsdTest, formEntryXsdDev, confFormEntryXsd, confFormEntryXsdUAT, confFormEntryXsdTest, confFormEntryXsdDev]

    unexpectedXsds = set(doc.modelDocument.uri
                         for doc, referencingDoc in modelXbrl.modelDocument.referencesDocument.items()
                         if "href" in referencingDoc.referenceTypes
                         if doc.modelDocument.uri not in allFormEntryXsd)
    if unexpectedXsds:
        modelXbrl.error("FERC.22.00",
                        _("The instance document contained unexpected schema references %(schemaReferences)s."),
                        modelXbrl=modelXbrl, schemaReferences=", ".join(sorted(unexpectedXsds)))

    if contextIDs: # check if contextID is on any undefined facts
        for undefinedFact in modelXbrl.undefinedFacts:
            contextIDs.discard(undefinedFact.get("contextRef"))
        if contextIDs:
            modelXbrl.error("FERC.6.05.08",
                            _("The instance document contained context(s) %(contextIDs)s that was(were) not used in any fact."),
                            modelXbrl=modelXbrl, contextIDs=", ".join(str(c) for c in contextIDs))
    if precisionFacts:
            modelXbrl.error("FERC.6.05.17",
                _("The instance document contains elements using the precision attribute."),
                modelObject=precisionFacts)

    #6.5.14 facts without english text
    for keyNotDefaultLang, factNotDefaultLang in keysNotDefaultLang.items():
        anyDefaultLangFact = False
        for fact in factsForLang[keyNotDefaultLang]:
            if fact.xmlLang.lower() == requiredFactLang: #.startswith(factLangStartsWith):
                anyDefaultLangFact = True
                break
        if not anyDefaultLangFact:
            val.modelXbrl.error("FERC.6.05.14",
                _("Element %(fact)s in context %(contextID)s has text with xml:lang other than '%(lang2)s' (%(lang)s) without matching English text.  "),
                modelObject=factNotDefaultLang, fact=factNotDefaultLang.qname, contextID=factNotDefaultLang.contextID,
                lang=factNotDefaultLang.xmlLang, lang2=disclosureSystem.defaultXmlLang) # report lexical format default lang

    modelXbrl.profileActivity(_statusMsg, minTimeToShow=0.0)
    modelXbrl.modelManager.showStatus(None)


__pluginInfo__ = {
    # Do not use _( ) in pluginInfo itself (it is applied later, after loading
    'name': 'Validate FERC',
    'version': '1.0',
    'description': '''FERC (US) Validation.''',
    'license': 'Apache-2',
    'author': authorLabel,
    'copyright': copyrightLabel,
    'import': ('inlineXbrlDocumentSet', ), # import dependent modules
    # classes of mount points (required)
    'DisclosureSystem.Types': disclosureSystemTypes,
    'DisclosureSystem.ConfigURL': disclosureSystemConfigURL,
    'Validate.XBRL.Start': validateXbrlStart,
    'Validate.XBRL.Finally': validateXbrlFinally,
}



'''
See COPYRIGHT.md for copyright information.
Python version specific utilities

do not convert 3 to 2
'''
from __future__ import annotations

import fractions
import os
import subprocess
import sys
from collections import OrderedDict
from collections.abc import Iterable, Iterator, Mapping, MappingView, MutableSet, Set
from decimal import Decimal
from types import MappingProxyType
from typing import Any, Generic, TypeVar

from arelle.typing import OptionalString

STR_NUM_TYPES = (str, int, float, Decimal, fractions.Fraction)

# python 3 unquote, because py2 unquote doesn't do utf-8 correctly
def py3unquote(string, encoding='utf-8', errors='replace'):
    """Replace %xx escapes by their single-character equivalent. The optional
    encoding and errors parameters specify how to decode percent-encoded
    sequences into Unicode characters, as accepted by the bytes.decode()
    method.
    By default, percent-encoded sequences are decoded with UTF-8, and invalid
    sequences are replaced by a placeholder character.

    unquote('abc%20def') -> 'abc def'.
    """
    if string == '':
        return string
    res = string.split('%')
    if len(res) == 1:
        return string
    if encoding is None:
        encoding = 'utf-8'
    if errors is None:
        errors = 'replace'
    # pct_sequence: contiguous sequence of percent-encoded bytes, decoded
    pct_sequence = b''
    string = res[0]
    for item in res[1:]:
        try:
            if not item:
                raise ValueError
            pct_sequence += str(bytearray.fromhex(item[:2]))
            rest = item[2:]
            if not rest:
                # This segment was just a single percent-encoded character.
                # May be part of a sequence of code units, so delay decoding.
                # (Stored in pct_sequence).
                continue
        except ValueError:
            rest = '%' + item
        # Encountered non-percent-encoded characters. Flush the current
        # pct_sequence.
        string += pct_sequence.decode(encoding, errors) + rest
        pct_sequence = b''
    if pct_sequence:
        # Flush the final pct_sequence
        string += pct_sequence.decode(encoding, errors)
    return string

def pyTypeName(object):
    try:
        objectClass = object.__class__
        classModule = objectClass.__module__
        className = objectClass.__name__
        if classModule == 'builtins':
            return className
        fullname = classModule + '.' + className
        if fullname == 'arelle.ModelValue.DateTime':
            if object.dateOnly:
                fullname += '-dateOnly'
            else:
                fullname += '-dateTime'
        return fullname
    except:
        return str(type(object))

def pyNamedObject(name, *args, **kwargs):
    try:
        import builtins
        objectConstructor = builtins.__dict__[name]
        return objectConstructor(*args, **kwargs)
    except:
        return None

def lcStr(value): # lower case first letter of string
    if len(value):
        return value[0].lower() + value[1:]
    return str

def strTruncate(value, length) -> str:
    _s = str(value).strip()
    if len(_s) <= length:
        return _s
    return _s[0:length-3] + "..."


def normalizeSpace(s: OptionalString) -> OptionalString:
    if isinstance(s, str):
        return " ".join(s.split())
    return s

SEQUENCE_TYPES = (tuple,list,set,frozenset,MappingView)
def flattenSequence(x, sequence=None) -> list[Any]:
    if sequence is None:
        if not isinstance(x, SEQUENCE_TYPES):
            if x is None:
                return [] # none as atomic value is an empty sequence in xPath semantics
            return [x]
        sequence = []
    for el in x:
        if isinstance(el, SEQUENCE_TYPES):
            flattenSequence(el, sequence)
        else:
            sequence.append(el)
    return sequence

def flattenToSet(x, _set=None):
    if _set is None:
        if not isinstance(x, SEQUENCE_TYPES):
            if x is None:
                return set() # none as atomic value is an empty sequence in xPath semantics
            return {x}
        _set = set()
    for el in x:
        if isinstance(el, SEQUENCE_TYPES):
            flattenToSet(el, _set)
        else:
            _set.add(el)
    return _set

class attrdict(dict):
    """ utility to simulate an object with named fields from a dict """
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        self.__dict__ = self

class OrderedDefaultDict(OrderedDict):
    """ call with default factory and optional sorted initial entries
        e.g., OrderedDefaultDict(list, ((1,11),(2,22),...))
    """
    def __init__(self, *args):
        self.default_factory = None
        if len(args) > 0:
            # arg0 is default_factory
            self.default_factory = args[0]
        if len(args) > 1:
            # arg1 is initial contents
            super(OrderedDefaultDict, self).__init__(args[1])
        else:
            super(OrderedDefaultDict, self).__init__()

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        _missingValue = self.default_factory()
        self[key] = _missingValue
        return _missingValue


T = TypeVar('T')

class OrderedSet(MutableSet[T]):
    """
    OrderedSet implementation copied from Python recipe:
    https://code.activestate.com/recipes/576694/
    """

    def __init__(self, iterable: Iterable[T] | None = None) -> None:
        self.end: list[Any] = []
        end = self.end
        end += [None, end, end]            # sentinel node for doubly linked list
        self.map: dict[T, list[Any]] = {}  # key --> [key, prev, next]
        if iterable is not None:
            self.update(iterable)

    def __getitem__(self, index: int) -> T:
        if not isinstance(index, int):
            raise TypeError("Index must be an integer")
        if index < 0:
            index += len(self)
        if index < 0 or index >= len(self):
            raise IndexError("Index out of range")
        end = self.end
        curr = end[2]
        for _ in range(index):
            curr = curr[2]
        return curr[0]

    def __len__(self) -> int:
        return len(self.map)

    def __contains__(self, key: object) -> bool:
        return key in self.map

    def add(self, key: T) -> None:
        if key not in self.map:
            end = self.end
            curr = end[1]
            curr[2] = end[1] = self.map[key] = [key, curr, end]

    def update(self, other: Iterable[T]) -> None:
        for key in other:
            self.add(key)

    def discard(self, key: T) -> None:
        if key in self.map:
            key, prev, next = self.map.pop(key)
            prev[2] = next
            next[1] = prev

    def __iter__(self) -> Iterator[T]:
        end = self.end
        curr = end[2]
        while curr is not end:
            yield curr[0]
            curr = curr[2]

    def __reversed__(self) -> Iterator[T]:
        end = self.end
        curr = end[1]
        while curr is not end:
            yield curr[0]
            curr = curr[1]

    def pop(self, last: bool = True) -> T:
        if not self:
            raise KeyError('set is empty')
        key = self.end[1][0] if last else self.end[2][0]
        self.discard(key)
        return key

    def __repr__(self) -> str:
        if not self:
            return f'{self.__class__.__name__}()'
        return f'{self.__class__.__name__}({list(self)!r})'

    def __eq__(self, other: object) -> bool:
        if isinstance(other, OrderedSet):
            return len(self) == len(other) and list(self) == list(other)
        if isinstance(other, Iterable):
            return set(self) == set(other)
        return NotImplemented

class FrozenOrderedSet(Set[T]):
    """
    Like frozenset vs set, this is the immutable counterpart to OrderedSet.
    Maintains insertion order and provides set-like operations without mutation.
    """

    def __init__(self, iterable: Iterable[T] | None = None) -> None:
        if iterable is None:
            self._items: tuple[T, ...] = ()
            self._set: frozenset[T] = frozenset()
        else:
            unique_items = dict.fromkeys(iterable)
            self._items = tuple(unique_items.keys())
            self._set = frozenset(unique_items.keys())
        self._hash: int | None = None

    def __getitem__(self, index: int) -> T:
        return self._items[index]

    def __len__(self) -> int:
        return len(self._items)

    def __contains__(self, key: object) -> bool:
        return key in self._set

    def __iter__(self) -> Iterator[T]:
        return iter(self._items)

    def __reversed__(self) -> Iterator[T]:
        return reversed(self._items)

    def __repr__(self) -> str:
        if not self:
            return f'{self.__class__.__name__}()'
        return f'{self.__class__.__name__}({self._items!r})'

    def __eq__(self, other: object) -> bool:
        if isinstance(other, (FrozenOrderedSet, OrderedSet)):
            return len(self) == len(other) and list(self) == list(other)
        if isinstance(other, Iterable):
            return set(self) == set(other)
        return NotImplemented

    def __hash__(self) -> int:
        if self._hash is None:
            self._hash = hash(self._items)
        return self._hash

KT = TypeVar('KT')
VT = TypeVar('VT')


class FrozenDict(Generic[KT, VT], Mapping[KT, VT]):
    def __init__(self, data: Mapping[KT, VT] | None = None) -> None:
        self._dict: Mapping[KT, VT] = MappingProxyType(dict(data) if data is not None else dict())
        self._hash: int | None = None

    def __getitem__(self, key: KT) -> VT:
        return self._dict[key]

    def __iter__(self) -> Iterator[KT]:
        return iter(self._dict)

    def __len__(self) -> int:
        return len(self._dict)

    def __repr__(self) -> str:
        if not self:
            return f'{self.__class__.__name__}()'
        return f"{self.__class__.__name__}({self._dict})"

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, FrozenDict):
            return self._dict == other._dict
        if isinstance(other, Mapping):
            return self._dict == other
        return NotImplemented

    def __hash__(self) -> int:
        if self._hash is None:
            self._hash = hash(tuple(sorted(self._dict.items())))
        return self._hash


def Fraction(numerator,denominator=None):
    if denominator is None:
        if isinstance(numerator, (fractions.Fraction,str,Decimal)):
            return Fraction(numerator)
    elif isinstance(numerator, Decimal) and isinstance(denominator, Decimal):
        return Fraction(int(numerator), int(denominator))
    return Fraction(numerator, denominator)

def pyObjectSize(obj, seen=None):
    """Recursively finds size of objects"""
    size = sys.getsizeof(obj)
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)
    if isinstance(obj, dict):
        size += sum([pyObjectSize(v, seen) for v in obj.values()])
        size += sum([pyObjectSize(k, seen) for k in obj.keys()])
    elif hasattr(obj, '__dict__'):
        size += pyObjectSize(obj.__dict__, seen)
    elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
        size += sum([pyObjectSize(i, seen) for i in obj])
    return size


def tryRunCommand(*args: str) -> str | None:
    """
    Tries to return the results of the provided command.
    Returns stdout or None if the command exists with a non-zero code.
    """
    try:
        return subprocess.run(
            args,
            capture_output=True,
            check=True,
            text=True,
            # A call to get std handle throws an OSError if stdin is not specified when run on Windows as a service.
            stdin=subprocess.PIPE,
        ).stdout.strip()
    except (OSError, subprocess.SubprocessError):
        return None

def isLegacyAbs(path: str) -> bool:
    """
    Implements legacy behavior of os.path.isabs() prior to Python 3.13
    where True was returned for paths beginning with single slashes on Windows.
    """
    if os.path.isabs(path):
        return True
    from arelle.SystemInfo import PlatformOS
    if PlatformOS.getPlatformOS() == PlatformOS.WINDOWS:
        return path.startswith(("/", "\\"))
    return False
